# -*- coding: utf-8 -*-
#
# Fixtures for administration.
#
# ------------------------------------------------


# imports
# -------
import pytest
import factory
from flask import Flask, render_template, g
from flask_sqlalchemy import SQLAlchemy
from flask_authorize import Authorize, PermissionsMixin, AllowancesMixin, RestrictionsMixin
from flask_authorize.mixins import default_allowances, default_restrictions, default_permissions

from . import SANDBOX


# application
# -----------
class Config(object):
    ENV = 'testing'
    TESTING = True
    SQLALCHEMY_ECHO = False
    PROPAGATE_EXCEPTIONS = False
    SQLALCHEMY_DATABASE_URI = 'sqlite:///{}/app.db'.format(SANDBOX)
    AUTHORIZE_ALLOW_ANONYMOUS_ACTIONS = True


app = Flask(__name__)
app.config.from_object(Config)
db = SQLAlchemy(app)
authorize = Authorize(app, current_user=lambda: g.user)


# models
# ------
UserGroup = db.Table(
    'user_group', db.Model.metadata,
    db.Column('user_id', db.Integer, db.ForeignKey('users.id')),
    db.Column('group_id', db.Integer, db.ForeignKey('groups.id'))
)


UserRole = db.Table(
    'user_role', db.Model.metadata,
    db.Column('user_id', db.Integer, db.ForeignKey('users.id')),
    db.Column('role_id', db.Integer, db.ForeignKey('roles.id'))
)


class User(db.Model):
    __tablename__ = 'users'
    __check_access__ = False

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False, unique=True, index=True)
    roles = db.relationship('Role', secondary=UserRole)
    groups = db.relationship('Group', secondary=UserGroup)


class Group(db.Model, RestrictionsMixin):
    __tablename__ = 'groups'
    __check_access__ = False

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False, unique=True, index=True)
    desc = db.Column(db.String(255))
    users = db.relationship('User', secondary=UserGroup)


class Role(db.Model, AllowancesMixin):
    __tablename__ = 'roles'
    __check_access__ = False

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False, unique=True, index=True)
    desc = db.Column(db.String(255))
    users = db.relationship('User', secondary=UserRole)


class Article(db.Model, PermissionsMixin):
    __tablename__ = 'articles'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), index=True, nullable=False)


class Item(db.Model):
    __tablename__ = 'items'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), index=True, nullable=False)


# endpoints
# ---------
@app.route('/feed', methods=['GET'])
def feed():
    articles = Article.query.filter_by(name='Jinja Article').all()
    return render_template('feed.html', articles=articles)


# factories
# ---------
class GroupFactory(factory.alchemy.SQLAlchemyModelFactory):

    id = factory.Sequence(lambda x: x + 100)
    name = factory.Faker('name')

    class Meta:
        model = Group
        sqlalchemy_session = db.session
        sqlalchemy_session_persistence = 'commit'


class RoleFactory(factory.alchemy.SQLAlchemyModelFactory):

    id = factory.Sequence(lambda x: x + 100)
    name = factory.Faker('name')

    class Meta:
        model = Role
        sqlalchemy_session = db.session
        sqlalchemy_session_persistence = 'commit'


class UserFactory(factory.alchemy.SQLAlchemyModelFactory):

    id = factory.Sequence(lambda x: x + 100)
    name = factory.Faker('name')

    class Meta:
        model = User
        sqlalchemy_session = db.session
        sqlalchemy_session_persistence = 'commit'


class ArticleFactory(factory.alchemy.SQLAlchemyModelFactory):

    id = factory.Sequence(lambda x: x + 100)
    name = factory.Faker('name')
    owner = factory.SubFactory(UserFactory)
    permissions = factory.LazyFunction(default_permissions)

    class Meta:
        model = Article
        sqlalchemy_session = db.session
        sqlalchemy_session_persistence = 'commit'


class ItemFactory(factory.alchemy.SQLAlchemyModelFactory):

    id = factory.Sequence(lambda x: x + 100)
    name = factory.Faker('name')

    class Meta:
        model = Item
        sqlalchemy_session = db.session
        sqlalchemy_session_persistence = 'commit'


# fixtures
# --------
@pytest.fixture(scope='session')
def anonymous(client):
    yield UserFactory.create(
        name='anonymous'
    )


@pytest.fixture(scope='session')
def admin(client):
    group = GroupFactory.create(name='admins')
    role = RoleFactory.create(name='admins')
    yield UserFactory.create(
        name='admin',
        groups=[group],
        roles=[role]
    )


@pytest.fixture(scope='session')
def reader(client):
    group = GroupFactory.create(name='readers')
    role = RoleFactory.create(name='readers')
    yield UserFactory.create(
        name='reader',
        groups=[group],
        roles=[role]
    )


@pytest.fixture(scope='session')
def editor(client):
    group = GroupFactory.create(name='editors')
    role = RoleFactory.create(name='editors')
    yield UserFactory.create(
        name='editor',
        groups=[group],
        roles=[role]
    )


@pytest.fixture(scope='session')
def allowed(client):
    role = RoleFactory.create(
        name='allowed',
        allowances=default_allowances()
    )
    yield UserFactory.create(
        name='allowed',
        roles=[role]
    )


@pytest.fixture(scope='session')
def unallowed(client):
    role = RoleFactory.create(
        name='unallowed',
        allowances={}
    )
    yield UserFactory.create(
        name='unallowed',
        roles=[role]
    )


@pytest.fixture(scope='session')
def restricted(client):
    group = GroupFactory.create(
        name='restricted',
        restrictions=default_allowances()
    )
    yield UserFactory.create(
        name='restricted',
        groups=[group]
    )


@pytest.fixture(scope='session')
def unrestricted(client):
    group = GroupFactory.create(
        name='unrestricted',
        restrictions=default_restrictions()
    )
    yield UserFactory.create(
        name='unrestricted',
        groups=[group]
    )



# -*- coding: utf-8 -*-
#
# Database mixins
#
# ------------------------------------------------


# imports
# -------
import six
import re
import json
from flask import current_app
from werkzeug.exceptions import Unauthorized
from sqlalchemy import Column, ForeignKey
from sqlalchemy.types import Integer, Text
from sqlalchemy.orm import relationship
from sqlalchemy.sql import operators
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy import TypeDecorator, inspect, and_, or_


# types
# -----
class JSON(TypeDecorator):
    """
    SQLite, MySQL, and PostgreSQL compatible type
    for json column.
    """
    impl = Text

    @property
    def python_type(self):
        return object

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        try:
            return json.loads(value)
        except (ValueError, TypeError):
            return None


class PipedList(TypeDecorator):
    """
    SQLite, MySQL, and PostgreSQL compatible type
    for column that renders as list when referenced
    in python and a string where list entries are
    separated by pipes '|' when referenced in the
    database.
    """
    impl = Text

    @property
    def python_type(self):
        return object

    def coerce_compared_value(self, op, value):
        if op in (operators.like_op,
                  operators.notlike_op,
                  operators.contains_op):
            return Text()
        else:
            return self

    def process_bind_param(self, value, dialect):
        if not value:
            return None
        return '|'.join(value)

    def process_result_value(self, value, dialect):
        try:
            if not value:
                return []
            return value.split('|')
        except (ValueError, TypeError):
            return None


# helpers
# -------
MODELS = dict()


def gather_models():
    """
    Inspect sqlalchemy models from current context and set global
    dictionary to be used in url conversion.
    """
    global MODELS

    from flask import current_app
    if 'sqlalchemy' not in current_app.extensions:
        return
    check = current_app.config['AUTHORIZE_IGNORE_PROPERTY']

    # inspect current models and add to map
    db = current_app.extensions['sqlalchemy'].db
    for cls in db.Model._decl_class_registry.values():
        if isinstance(cls, type) and issubclass(cls, db.Model):
            if hasattr(cls, check) and not getattr(cls, check):
                continue
            MODELS[table_key(cls)] = cls
    return


def table_key(cls):
    """
    Parse table key from sqlalchemy class, based on user-specified
    configuration included for extension.
    """
    # class name
    if current_app.config['AUTHORIZE_MODEL_PARSER'] == 'class':
        return cls.__name__

    # lowercase name
    elif current_app.config['AUTHORIZE_MODEL_PARSER'] == 'lower':
        return cls.__name__.lower()

    # snake_case name
    elif current_app.config['AUTHORIZE_MODEL_PARSER'] == 'snake':
        words = re.findall(r'([A-Z][0-9a-z]+)', cls.__name__)
        if len(words) > 1:
            return '_'.join(map(lambda x: x.lower(), words))

    # table name
    elif current_app.config['AUTHORIZE_MODEL_PARSER'] == 'table':
        mapper = inspect(cls)
        return mapper.tables[0].name


def default_permissions_factory(name):
    """
    Factory for returning default permissions based on name.
    """
    def _(cls=None):
        perms = default_permissions(cls)
        return perms.get(name, [])
    return _


def default_permissions(cls=None):
    """
    Return default permissions for model, falling
    back to app configuration if no default permission
    is explicitly set.
    """
    if cls is None or cls.__permissions__ is None:
        return current_app.config['AUTHORIZE_DEFAULT_PERMISSIONS']
    elif isinstance(cls._permissions__, int):
        return parse_permission_set(cls.__permissions__)
    elif isinstance(cls.__permissions__, dict):
        return cls.__permissions__


def default_allowances(cls=None):
    """
    Return default permissions for model, falling
    back to app configuration if no default permission
    is explicitly set.
    """
    # if necessary, gather database models to create default
    global MODELS
    if not MODELS:
        gather_models()

    # configure defaults
    default = {
        key: current_app.config['AUTHORIZE_DEFAULT_ALLOWANCES']
        for key in MODELS
    }

    # if called directly, return the defaults
    if cls is None:
        return default

    # if allowances are explicitly set to something else, use them
    if isinstance(cls.__allowances__, dict):
        return cls.__allowances__

    # otherwise, use defaults
    return default


def default_restrictions(cls=None):
    """
    Return default permissions for model, falling
    back to app configuration if no default permission
    is explicitly set.
    """
    # if necessary, gather database models to create default
    global MODELS
    if not MODELS:
        gather_models()

    # configure defaults
    default = {
        key: current_app.config['AUTHORIZE_DEFAULT_RESTRICTIONS']
        for key in MODELS
    }

    # if called directly, return the defaults
    if cls is None:
        return default

    # if set to fail safe, use that
    if cls.__restrictions__ == '*' or cls.__restrictions__ is True:
        return {
            key: current_app.config['AUTHORIZE_DEFAULT_ACTIONS']
            for key in MODELS
        }

    # overwrite specified allowances
    if isinstance(cls.__restrictions__, dict):
        default.update(cls.__restrictions__)
    return default


def permission_list(number):
    """
    Generate permission list from numeric input.
    """
    if isinstance(number, six.string_types) and len(number) == 1:
        number = int(number)
    if not isinstance(number, int):
        return number

    ret = []
    for mask, name in zip([1, 2, 4], ['delete', 'read', 'update']):
        if number & mask:
            ret.append(name)
    return ret


def parse_permission_set(number):
    """
    Parse numeric permissions and return dictionary with
    explicit permission scheme. Note that this method
    does not account for custom content permissions.
    """
    if isinstance(number, six.string_types) and len(number) == 3:
        number = int(number)
    if not isinstance(number, int):
        return number

    # check validity of input
    digits = len(str(number))
    if digits > 3:
        raise AssertionError('Invalid permissions: {}'.format(number))

    # gather permissions
    result = {}
    for digit, check in zip([0, 1, 2], ['other', 'group', 'owner']):
        perm = int(number) // 10 ** digit % 10
        result[check] = permission_list(perm)
    return result


# permissions mixins
# ------------------
class BasePermissionsMixin(object):
    """
    Abstract base class for enabling common functionality
    across various optional permission schemes.
    """
    __permissions__ = None

    @declared_attr
    def other_permissions(cls):
        return Column(PipedList, default=default_permissions_factory('other'))

    @classmethod
    def authorized(cls, check):
        """
        Query operator for permissions mixins. This operator
        can be used in SQLAlchemy query statements, and will
        automatically decorate queries with appropriate owner/group
        and permissionc checks.

        Arguments:
            check (str): Permission to authorize (i.e. read, update)

        Examples:

            Query all articles where the current user is read-authorized:

            .. code-block:: python

                Article.query.filter(Article.authorized('read')).all()


            Query by multiple parameters, including authorization:

            .. code-block:: python

                Article.query.filter(or_(
                    Article.name.contains('open article'),
                    Article.authorized('read')
                ))
        """
        from .plugin import CURRENT_USER
        current_user = CURRENT_USER()
        clauses = [
            cls.other_permissions.contains(check),
        ]
        if hasattr(current_user, 'id'):
            if hasattr(cls, 'owner_id'):
                clauses.append(and_(
                    current_user.id == cls.owner_id,
                    cls.owner_permissions.contains(check)
                ))
            if hasattr(cls, 'group_id') and hasattr(current_user, 'groups'):
                clauses.append(and_(
                    cls.group_id.in_([x.id for x in current_user.groups]),
                    cls.group_permissions.contains(check)
                ))
        return or_(*clauses)

    @property
    def permissions(self):
        """
        Proxy for interacting with permissions dictionary.
        """
        result = {}
        for name in ['owner', 'group', 'other']:
            prop = name + '_permissions'
            if hasattr(self, prop):
                result[name] = getattr(self, prop)
        return result

    @permissions.setter
    def permissions(self, value):
        """
        Setter for permissions dictionary proxy.
        """
        for name in ['owner', 'group', 'other']:
            if name not in value:
                continue
            prop = name + '_permissions'
            if hasattr(self, prop):
                setattr(self, prop, value[name])
        return

    def set_permissions(self, *args, **kwargs):
        """
        Set permissions explicitly for ACL-enforced content.
        """
        if 'authorize' in current_app.extensions:
            authorize = current_app.extensions['authorize']
            if not authorize.update(self):
                raise Unauthorized

        # handle numeric permission scheme
        if len(args):
            perms = parse_permission_set(args[0])
            kwargs.update(perms)

        # set internal permissions object
        permissions = self.permissions.copy()
        permissions.update(kwargs)
        self.permissions = permissions
        return self


class OwnerMixin(object):
    """
    Mixin providing owner-related database properties
    for object, in the context of enforcing permissions.
    """
    @declared_attr
    def owner_id(cls):
        return Column(Integer, ForeignKey('users.id'))

    @declared_attr
    def owner(cls):
        return relationship('User')

    @declared_attr
    def owner_permissions(cls):
        return Column(PipedList, default=default_permissions_factory('owner'))


class OwnerPermissionsMixin(BasePermissionsMixin, OwnerMixin):
    pass


class GroupMixin(object):
    """
    Mixin providing group-related database properties
    for object, in the context of enforcing permissions.
    """
    @declared_attr
    def group_id(cls):
        return Column(Integer, ForeignKey('groups.id'))

    @declared_attr
    def group(cls):
        return relationship('Group')

    @declared_attr
    def group_permissions(cls):
        return Column(PipedList, default=default_permissions_factory('group'))


class GroupPermissionsMixin(BasePermissionsMixin, GroupMixin):
    pass


# class MultiGroupMixin(object):
#     """
#     Mixin providing groups-related database properties
#     for object, in the context of enforcing permissions.

#     .. note:: NEEDS MORE DOCUMENTATION AND EXAMPLES

#     .. note:: NEED TO FIGURE OUT HOW TO AUTOMATICALLY CREATE MAPPING TABLE
#     """
#     @declared_attr
#     def groups(cls):
#         return relationship('Group', backref=backref(
#             'articles', cascade="all, delete-orphan",
#         ))


# class MultiGroupPermissionsMixin(BasePermissionsMixin, MultiGroupMixin):
#     pass


class PermissionsMixin(BasePermissionsMixin, OwnerMixin, GroupMixin):
    """
    Mixin providing owner and group-related database properties
    for object, in the context of enforcing permissions.
    """
    pass


# OwnerGroupPermissionsMixin = PermissionsMixin


# class ComplexPermissionsMixin(BasePermissionsMixin, OwnerMixin, MultiGroupMixin):
#     """
#     Mixin providing owner and multi-group-related database
#     properties for object, in the context of enforcing permissions.

#     .. note:: NEEDS MORE DOCUMENTATION AND EXAMPLES
#     """
#     pass


# OwnerGroupsPermissionMixin = ComplexPermissionsMixin


# rbac mixins
# -----------
class RestrictionsMixin(object):
    """
    Mixin providing group or role based access control.
    """
    __restrictions__ = dict()

    @declared_attr
    def restrictions(cls):
        return Column(JSON, default=default_restrictions)

    def set_restrictions(self, **kwargs):
        # handle numeric permission scheme
        for key in kwargs:
            kwargs[key] = permission_list(kwargs[key])

        # set internal restrictions object
        restrictions = self.restrictions.copy()
        restrictions.update(kwargs)
        self.restrictions = restrictions
        return self


class AllowancesMixin(object):
    """
    Mixin providing group or role based access control.
    """
    __allowances__ = '*'

    @declared_attr
    def allowances(cls):
        return Column(JSON, default=default_allowances)

    def set_allowances(self, **kwargs):
        # handle numeric permission scheme
        for key in kwargs:
            kwargs[key] = permission_list(kwargs[key])

        # set internal restrictions object
        allowances = self.allowances.copy()
        allowances.update(kwargs)
        self.allowances = allowances
        return self



# -*- coding: utf-8 -*-

__pkg__ = 'Flask-Authorize'
__url__ = 'https://github.com/bprinty/Flask-Authorize'
__info__ = 'Flask plugin for content authorization and access control'
__author__ = 'Blake Printy'
__email__ = 'bprinty@gmail.com'
__version__ = '0.1.10'


from .mixins import RestrictionsMixin           ## noqa
from .mixins import AllowancesMixin             ## noqa

from .mixins import OwnerPermissionsMixin       ## noqa
from .mixins import GroupPermissionsMixin       ## noqa
from .mixins import PermissionsMixin            ## noqa

from .mixins import default_permissions         ## noqa
from .mixins import default_allowances          ## noqa
from .mixins import default_restrictions        ## noqa

from .plugin import Authorize                   ## noqa



# -*- coding: utf-8 -*-
#
# Testing for access control list authorization.
#
# ------------------------------------------------


# imports
# -------
from sqlalchemy import and_
from flask import g
from .fixtures import authorize, Article, ArticleFactory


# helpers
# -------
def query(name, check):
    return Article.query.filter(and_(
        Article.name.contains(name),
        Article.authorized(check)
    )).all()


# session
# -------
class TestOtherPermissions(object):

    def test_other_delete(self, client, reader, editor):

        # other open read permissions
        g.user = None
        article = ArticleFactory.create(
            name='Other Delete Open Article',
            owner=editor,
            group=editor.groups[0]
        ).set_permissions('001')

        g.user = reader
        assert authorize.delete(article)
        assert query(article.name, 'delete')

        # other closed read permissions
        g.user = None
        article = ArticleFactory.create(
            name='Other Delete Closed Article',
            owner=editor,
            group=editor.groups[0]
        ).set_permissions('770')

        g.user = reader
        assert not authorize.delete(article)
        assert not query(article.name, 'delete')
        return

    def test_other_read(self, client, reader, editor):

        # other open read permissions
        g.user = None
        article = ArticleFactory.create(
            name='Other Read Open Article',
            owner=editor,
            group=editor.groups[0]
        ).set_permissions('002')

        g.user = reader
        assert authorize.read(article)
        assert query(article.name, 'read')

        # other closed read permissions
        g.user = None
        article = ArticleFactory.create(
            name='Other Read Closed Article',
            owner=editor,
            group=editor.groups[0]
        ).set_permissions('660')

        g.user = reader
        assert not authorize.read(article)
        assert not query(article.name, 'read')
        return

    def test_other_update(self, reader, editor):

        # other open update permissions
        g.user = None
        article = ArticleFactory.create(
            name='Other Write Open Article',
            owner=editor,
            group=editor.groups[0]
        ).set_permissions('004')

        g.user = reader
        assert authorize.update(article)
        assert query(article.name, 'update')

        # other closed update permissions
        g.user = None
        article = ArticleFactory.create(
            name='Other Write Closed Article',
            owner=editor,
            group=editor.groups[0]
        ).set_permissions('662')
        g.user = reader
        assert not authorize.update(article)
        assert not query(article.name, 'update')
        return

    def test_other_custom(self, reader, editor):
        # other open custom permissions
        g.user = None
        article = ArticleFactory.create(
            name='Other Custom Open Article',
            owner=editor,
            group=editor.groups[0]
        ).set_permissions(other=['custom'])

        g.user = reader
        assert authorize.custom(article)
        assert query(article.name, 'custom')

        # other closed custom permissions
        g.user = None
        article = ArticleFactory.create(
            name='Other Custom Closed Article',
            owner=editor,
            group=editor.groups[0]
        )
        g.user = reader
        assert not authorize.custom(article)
        assert not query(article.name, 'custom')
        return


class TestOwnerPermissions(object):

    def test_owner_delete(self, client, reader, editor):
        g.user = reader

        # other open read permissions
        article = ArticleFactory.create(
            name='Owner Delete Open Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('100')
        assert authorize.delete(article)
        assert query(article.name, 'delete')

        # other closed read permissions
        article = ArticleFactory.create(
            name='Owner Delete Closed Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('070')
        assert not authorize.delete(article)
        assert not query(article.name, 'delete')
        return

    def test_owner_read(self, client, reader, editor):
        g.user = reader

        # other open read permissions
        article = ArticleFactory.create(
            name='Owner Read Open Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('200')
        assert authorize.read(article)
        assert query(article.name, 'read')

        # other closed read permissions
        article = ArticleFactory.create(
            name='Owner Read Closed Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('170')
        assert not authorize.read(article)
        assert not query(article.name, 'read')
        return

    def test_owner_update(self, reader, editor):
        g.user = reader

        # other open update permissions
        article = ArticleFactory.create(
            name='Owner Write Open Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('400')
        assert authorize.update(article)
        assert query(article.name, 'update')

        # other closed update permissions
        article = ArticleFactory.create(
            name='Owner Write Closed Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('270')
        assert not authorize.update(article)
        assert not query(article.name, 'update')
        return

    def test_owner_custom(self, reader, editor):
        g.user = reader

        # other open update permissions
        article = ArticleFactory.create(
            name='Owner Custom Open Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions(owner=['custom'])
        assert authorize.custom(article)
        assert query(article.name, 'custom')

        # other closed update permissions
        article = ArticleFactory.create(
            name='Owner Custom Closed Article',
            owner=reader,
            group=editor.groups[0]
        )
        assert not authorize.custom(article)
        assert not query(article.name, 'custom')
        return


class TestGroupPermissions(object):

    def test_group_delete(self, client, reader, editor):
        g.user = editor

        # other open read permissions
        article = ArticleFactory.create(
            name='Group Delete Open Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('010')
        assert authorize.delete(article)
        assert query(article.name, 'delete')

        # other closed read permissions
        article = ArticleFactory.create(
            name='Group Delete Closed Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('700')
        assert not authorize.delete(article)
        assert not query(article.name, 'delete')
        return

    def test_group_read(self, client, reader, editor):
        g.user = editor

        # other open read permissions
        article = ArticleFactory.create(
            name='Group Read Open Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('020')
        assert authorize.read(article)
        assert query(article.name, 'read')

        # other closed read permissions
        article = ArticleFactory.create(
            name='Group Read Closed Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('710')
        assert not authorize.read(article)
        assert not query(article.name, 'read')
        return

    def test_group_update(self, reader, editor):
        g.user = editor

        # other open update permissions
        article = ArticleFactory.create(
            name='Group Write Open Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('040')
        assert authorize.update(article)
        assert query(article.name, 'update')

        # other closed update permissions
        article = ArticleFactory.create(
            name='Group Write Closed Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('720')
        assert not authorize.update(article)
        assert not query(article.name, 'update')
        return

    def test_group_custom(self, reader, editor):
        g.user = editor

        # other open update permissions
        article = ArticleFactory.create(
            name='Group Write Open Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions(group=['custom'])
        assert authorize.custom(article)
        assert query(article.name, 'custom')

        # other closed update permissions
        article = ArticleFactory.create(
            name='Group Write Closed Article',
            owner=reader,
            group=editor.groups[0]
        )
        assert not authorize.custom(article)
        assert not query(article.name, 'custom')
        return



# -*- coding: utf-8 -*-
#
# Testing for role- and group-based access control.
#
# ------------------------------------------------


# imports
# -------
import pytest
from flask import g
from werkzeug.exceptions import Unauthorized
from sqlalchemy import and_, or_

from .fixtures import authorize, Article, ArticleFactory


# authorizers
# -----------
@authorize.read
@authorize.create(Article)
@authorize.in_group('admins')
def in_group_or_read_and_create(article):
    pass


@authorize.read
@authorize.has_role('admins')
def has_role_or_read(article):
    pass


# tests
# -----
class TestDefaults(object):

    def test_defaults(self, client):
        from flask_authorize import default_permissions
        assert default_permissions() == dict(
            owner=['delete', 'read', 'update'],
            group=['read', 'update'],
            other=['read']
        )

        from flask_authorize import default_allowances
        assert default_allowances() == dict(
            articles=['create', 'delete', 'read', 'update'],
            items=['create', 'delete', 'read', 'update'],
        )

        from flask_authorize import default_restrictions
        assert default_restrictions() == dict(
            articles=[],
            items=[]
        )
        return


class TestIntegration(object):

    def test_in_group_or_read_and_create(self, client, reader, editor, admin, anonymous):
        g.user = None
        article = ArticleFactory.create(
            name='Complex Article 1',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('770')

        # no errors
        g.user = reader
        in_group_or_read_and_create(article)

        g.user = editor
        in_group_or_read_and_create(article)

        g.user = admin
        in_group_or_read_and_create(article)

        # errors
        g.user = anonymous
        with pytest.raises(Unauthorized):
            in_group_or_read_and_create(article)
        return

    def test_has_role_or_read(self, client, reader, editor, admin, anonymous):
        g.user = None
        article = ArticleFactory.create(
            name='Complex Article 2',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('700')

        # no errors
        g.user = reader
        has_role_or_read(article)

        g.user = admin
        has_role_or_read(article)

        # errors
        g.user = anonymous
        with pytest.raises(Unauthorized):
            has_role_or_read(article)
        return

    def test_multiple_permissions(self, client, reader, editor):
        g.user = reader
        allow = ArticleFactory.create(
            name='Multiple Permissions Open Article',
            owner=reader,
            group=reader.groups[0]
        ).set_permissions('777')
        deny = ArticleFactory.create(
            name='Multiple Permissions Closed Article',
            owner=reader,
            group=reader.groups[0]
        ).set_permissions('000')

        g.user = reader
        assert authorize.delete(allow)
        assert not authorize.delete(deny)
        assert not authorize.delete(allow, deny)

        assert authorize.read(allow)
        assert not authorize.read(deny)
        assert not authorize.read(allow, deny)

        assert authorize.update(allow)
        assert not authorize.update(deny)
        assert not authorize.update(allow, deny)
        return


class TestQueryFilters(object):

    def test_query_filter_operations(self, reader, editor, anonymous):
        g.user = None
        article = ArticleFactory.create(
            name='Query Filter Article',
            owner=reader,
            group=editor.groups[0]
        ).set_permissions('770')

        # simple and_ filter
        g.user = reader
        articles = Article.query.filter(and_(
                Article.name == article.name,
                Article.authorized('read')
            )
        ).all()
        assert articles

        # complex and_/or_ operator
        g.user = editor
        articles = Article.query.filter(and_(
                Article.name == article.name,
                or_(
                    Article.authorized('read'),
                    Article.authorized('update')
                )
            )
        ).all()
        assert articles

        # and_ filter with negative query
        g.user = anonymous
        articles = Article.query.filter(and_(
                Article.name == article.name,
                Article.authorized('read')
            )
        ).all()
        assert not articles
        return



# -*- coding: utf-8 -*-
#
# Plugin Setup
#
# ------------------------------------------------


# imports
# -------
import six
import types
from functools import wraps
from flask import current_app
from werkzeug.exceptions import Unauthorized

from .mixins import default_permissions, default_allowances, table_key


# constants
# ---------
AUTHORIZE_CACHE = dict()
CURRENT_USER = None
EXCEPTION = None
STRICT = True


# default customizations
# ----------------------
def flask_login_current_user():
    try:
        from flask_login import current_user
        user = current_user
    except ImportError:
        raise AssertionError(
            'Error: Flask-Authorize requires that either '
            'Flask-Login is used or that `user` is '
            'specified to authorization method')
    return user


# plugin
# ------
class Authorize(object):
    """
    Plugin for updating flask functions to handle class-based URL
    routing.
    """

    def __init__(self, app=None, current_user=flask_login_current_user, exception=Unauthorized, strict=True):
        if app is not None:
            self.init_app(app)

        # set current user function
        if current_user is not None:
            if not callable(current_user):
                raise AssertionError('Error: `current_user` input must be callable.')
            global CURRENT_USER
            CURRENT_USER = current_user

        # set default exception
        if exception is not None:
            if not isinstance(exception, type):
                raise AssertionError('Error: `exception` input must be Exception type')
            global EXCEPTION
            EXCEPTION = Unauthorized

        # set mode
        global STRICT
        STRICT = strict
        return

    def init_app(self, app):
        # settings
        app.config.setdefault('AUTHORIZE_DEFAULT_PERMISSIONS', dict(
            owner=['delete', 'read', 'update'],
            group=['read', 'update'],
            other=['read']
        ))
        app.config.setdefault('AUTHORIZE_DEFAULT_ACTIONS', ['create', 'delete', 'read', 'update'])
        app.config.setdefault('AUTHORIZE_DEFAULT_RESTRICTIONS', [])
        app.config.setdefault('AUTHORIZE_DEFAULT_ALLOWANCES', app.config['AUTHORIZE_DEFAULT_ACTIONS'])
        app.config.setdefault('AUTHORIZE_MODEL_PARSER', 'table')
        app.config.setdefault('AUTHORIZE_IGNORE_PROPERTY', '__check_access__')
        app.config.setdefault('AUTHORIZE_ALLOW_ANONYMOUS_ACTIONS', False)
        app.config.setdefault('AUTHORIZE_DISABLE_JINJA', False)

        # add to extensions dict for access
        app.extensions['authorize'] = self
        self.app = app

        # add authorize decorator to jinja context
        if not app.config['AUTHORIZE_DISABLE_JINJA']:
            @app.context_processor
            def inject_authorize():
                return dict(authorize=self)
        return

    def __getattr__(self, key):
        return Authorizer(permission=key)

    @property
    def delete(self):
        return Authorizer(permission='delete')

    @property
    def read(self):
        return Authorizer(permission='read')

    @property
    def update(self):
        return Authorizer(permission='update')

    def create(self, *args):
        return Authorizer(create=args)

    def has_role(self, *args):
        return Authorizer(has_role=args)

    def in_group(self, *args):
        return Authorizer(in_group=args)


# helpers
# -------
def has_permission(expected, actual):
    """
    Check if singular set of expected/actual
    permissions are appropriate.
    """
    x = set(expected).intersection(actual)
    return len(x) == len(expected)


def user_has_role(user, roles):
    """
    Check if specified user has one of the specified roles.
    """
    global STRICT
    if not hasattr(user, 'roles'):
        return False
    for role in user.roles:
        if hasattr(role, 'name'):
            check = role.name
        elif not STRICT:
            check = str(role)
        else:
            raise AssertionError('`Role` model has no `name` property for checking membership.')
        if check in roles:
            return True
    return False


def user_in_group(user, groups):
    """
    Check if specified user is in one of the specified groups.
    """
    global STRICT
    if not hasattr(user, 'groups'):
        return False
    for group in user.groups:
        if hasattr(group, 'name'):
            check = group.name
        elif not STRICT:
            check = str(group)
        else:
            raise AssertionError('`Group` model has no `name` property for checking membership.')
        if check in groups:
            return True
    return False


def user_is_restricted(user, operation, obj):
    if isinstance(obj, six.string_types):
        key = obj
    elif isinstance(obj, type):
        key = table_key(obj)
    else:
        key = table_key(obj.__class__)

    # gather credentials to check
    credentials = []
    if hasattr(user, 'roles'):
        credentials.extend(user.roles)
    if hasattr(user, 'groups'):
        credentials.extend(user.groups)
    if not len(credentials):
        return False

    # check all credentials
    for cred in credentials:
        if hasattr(cred, 'restrictions') and cred.restrictions is not None:
            check = set(cred.restrictions.get(key, []))
            if len(check.intersection(operation)):
                return True
    return False


def user_is_allowed(user, operation, obj):
    if isinstance(obj, six.string_types):
        key = obj
    elif isinstance(obj, type):
        key = table_key(obj)
    else:
        key = table_key(obj.__class__)

    # gather credentials to check
    credentials = []
    if hasattr(user, 'roles'):
        credentials.extend(user.roles)
    if hasattr(user, 'groups'):
        credentials.extend(user.groups)
    if not len(credentials):
        return True

    # gather allowances from credentials
    allowances, default = [], default_allowances()
    for cred in credentials:

        # if not restricting allowances on one
        # of the credentials, it's allowed
        if not hasattr(cred, 'allowances'):
            return True
        if cred.allowances is None:
            return True

        allowances.extend(cred.allowances.get(key, default))

    # check allowances
    check = set(allowances).intersection(operation)
    if len(check) == len(operation):
        return True

    return False


# processor
# ---------
class Authorizer(object):
    """
    Decorator for authorizing the ability of the current
    user to perform actions on various models.

    .. code-block:: python

        @app.route('/users/<id(User):user>', method=['GET'])
        @authorize.read
        def get_user(user):
            return

        @app.route('/users/<id(User):user>', method=['PUT'])
        @authorize.update
        @authorize.has_role('user-updators')
        def update_user(user):
            return

        @app.route('/users/<ident>', method=['PUT'])
        def update_user(ident):
            user = User.get(id=ident)
            if not authorize.update(user) or not authorize.role('test-role'):
                raise Unauthorized
            return


    """

    def __init__(self, permission=None, has_role=None, in_group=None, create=None):
        def _(arg):
            if arg is None:
                arg = []
            if not isinstance(arg, (list, tuple)):
                arg = [arg]
            return list(arg)

        self.permission = _(permission)
        self.has_role = _(has_role)
        self.in_group = _(in_group)
        self.create = _(create)
        return

    def __bool__(self):
        """
        Proxy for doing conditional on functional methods (i.e. create)
        """
        return self.allowed()

    def __call__(self, *cargs, **ckwargs):

        # dispatch on whether or not being used as decorator
        if not len(cargs):
            raise AssertionError('Authorizer needs to be passed function for decoration or objects to authorize.')
        if not isinstance(cargs[0], types.FunctionType):
            return self.allowed(*cargs, user=ckwargs.get('user'))

        # allow for duplicate decorations on functions
        func = cargs[0]
        if func.__name__ not in AUTHORIZE_CACHE:
            AUTHORIZE_CACHE[func.__name__] = self
        else:
            original = AUTHORIZE_CACHE[func.__name__]
            updated = Authorizer(
                permission=original.permission + self.permission,
                has_role=original.has_role + self.has_role,
                in_group=original.in_group + self.in_group,
                create=original.create + self.create
            )
            AUTHORIZE_CACHE[func.__name__] = updated
            del original
            return func

        @wraps(func)
        def inner(*args, **kwargs):

            # gather all items to check authorization for
            check = list(args) + list(kwargs.values())

            # check if authorized
            auth = AUTHORIZE_CACHE[func.__name__]
            if not auth.allowed(*check):
                raise EXCEPTION

            return func(*args, **kwargs)

        return inner

    def allowed(self, *args, **kwargs):

        # look to flask-login for current user
        user = kwargs.get('user')
        if user is None:
            user = CURRENT_USER()

        # otherwise, use current user method
        elif isinstance(user, types.FunctionType):
            user = user()

        # don't allow anything for anonymous users
        if user is None:
            return current_app.config['AUTHORIZE_ALLOW_ANONYMOUS_ACTIONS']

        # authorize if user has relevant role
        if len(self.has_role):
            if user_has_role(user, self.has_role):
                return True
            elif not len(self.permission) and not len(self.create):
                return False

        # authorize if user has relevant group
        if len(self.in_group):
            if user_in_group(user, self.in_group):
                return True
            elif not len(self.permission) and not len(self.create):
                return False

        # authorize create privileges based on access
        if len(self.create):
            for model in self.create:
                if user_is_restricted(user, ['create'], model) or \
                   not user_is_allowed(user, ['create'], model):
                    return False

        # return if no additional permission check needed
        if len(self.permission) == 0:
            return True

        # check permissions on individual instances - all objects
        # must have authorization to proceed.
        operation = set(self.permission)
        for arg in args:

            if not isinstance(arg.__class__, six.class_types):
                continue

            # check role restrictions/allowances
            if user_is_restricted(user, operation, arg):
                return False

            if not user_is_allowed(user, operation, arg):
                return False

            # only check permissions for items that have set permissions
            check = current_app.config['AUTHORIZE_IGNORE_PROPERTY']
            if hasattr(arg, check) and not getattr(arg, check):
                continue
            if not hasattr(arg, 'permissions'):
                continue

            # check other permissions
            check = arg.permissions.get('other', [])
            permitted = has_permission(operation, check)

            # check user permissions
            if hasattr(arg, 'owner'):
                if arg.owner == user:
                    check = arg.permissions.get('owner', [])
                    permitted |= has_permission(operation, check)

            # check group permissions
            if hasattr(arg, 'group'):
                if hasattr(user, 'groups'):
                    if arg.group in user.groups:
                        check = arg.permissions.get('group', [])
                        permitted |= has_permission(operation, check)

            if not permitted:
                return False

        return True



# -*- coding: utf-8 -*-
#
# Testing for role- and group-based access control.
#
# ------------------------------------------------


# imports
# -------
from flask import g

from .fixtures import Article, ArticleFactory, Item, ItemFactory, authorize


# tests
# -----
class TestAccessControl(object):

    def test_allowances(self, client, reader, allowed, unallowed):
        g.user = None
        article = ArticleFactory.create(
            name='Allowances Open Article',
            owner=reader,
            group=reader.groups[0]
        ).set_permissions('777')
        item = ItemFactory.create()

        g.user = allowed
        assert authorize.read(article)
        assert authorize.create(Article)
        assert authorize.read(item)
        assert authorize.create(item)

        g.user = unallowed
        assert not authorize.read(article)
        assert not authorize.create(Article)
        assert not authorize.item(item)
        assert not authorize.item(Item)
        return

    def test_restrictions(self, client, reader, restricted, unrestricted):
        g.user = None
        article = ArticleFactory.create(
            name='Restrictions Open Article',
            owner=reader,
            group=reader.groups[0]
        ).set_permissions('777')
        item = ItemFactory.create()

        g.user = unrestricted
        assert authorize.read(article)
        assert authorize.create(Article)
        assert authorize.read(item)
        assert authorize.create(Item)

        g.user = restricted
        assert not authorize.read(article)
        assert not authorize.create(Article)
        assert not authorize.read(item)
        assert not authorize.create(Item)
        return


class TestCredentials(object):

    def test_in_group(self, client, admin, reader, editor):
        g.user = reader
        assert authorize.in_group('readers')
        assert not authorize.in_group('editors')

        g.user = editor
        assert authorize.in_group('editors')
        assert not authorize.in_group('readers')
        return

    def test_has_role(self, client, reader, editor):
        g.user = reader
        assert authorize.has_role('readers')
        assert not authorize.has_role('editors')

        g.user = editor
        assert authorize.in_group('editors')
        assert not authorize.in_group('readers')
        return



#!/usr/bin/env python
#
# File used for debugging issues.
#
# -----------------------------------------


# imports
# -------
import os
from flask import Flask, render_template, g, jsonify, request
from flask_sqlalchemy import SQLAlchemy
from flask_login import LoginManager, UserMixin, login_user, login_required, current_user
from flask_authorize import Authorize, PermissionsMixin, AllowancesMixin, RestrictionsMixin
from flask_authorize.mixins import default_allowances, default_restrictions, default_permissions


# application
# -----------
class Config(object):
    ENV = 'testing'
    TESTING = True
    SQLALCHEMY_ECHO = False
    SECRET_KEY = '67yuhj'
    PROPAGATE_EXCEPTIONS = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_DATABASE_URI = 'sqlite:///{}/app.db'.format(os.getcwd())
    AUTHORIZE_ALLOW_ANONYMOUS_ACTIONS = True


app = Flask(__name__)
app.config.from_object(Config)
db = SQLAlchemy(app)
login_manager = LoginManager(app)
authorize = Authorize(app)


# models
# ------
UserGroup = db.Table(
    'user_group', db.Model.metadata,
    db.Column('user_id', db.Integer, db.ForeignKey('users.id')),
    db.Column('group_id', db.Integer, db.ForeignKey('groups.id'))
)


UserRole = db.Table(
    'user_role', db.Model.metadata,
    db.Column('user_id', db.Integer, db.ForeignKey('users.id')),
    db.Column('role_id', db.Integer, db.ForeignKey('roles.id'))
)


class User(db.Model, UserMixin):
    __tablename__ = 'users'
    # __check_access__ = False

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False, unique=True, index=True)
    roles = db.relationship('Role', secondary=UserRole)
    groups = db.relationship('Group', secondary=UserGroup)


class Group(db.Model, RestrictionsMixin):
    __tablename__ = 'groups'
    # __check_access__ = False

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False, unique=True, index=True)
    users = db.relationship('User', secondary=UserGroup)


class Role(db.Model, AllowancesMixin):
    __tablename__ = 'roles'
    # __check_access__ = False

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), nullable=False, unique=True, index=True)
    users = db.relationship('User', secondary=UserRole)


class Item(db.Model):
    __tablename__ = 'items'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(255), index=True, nullable=False)


# routes
# ------
@app.route('/register')
def register():
    """
    Register new user with specified roles and groups.
    """
    if 'name' not in request.args:
        raise AssertionError('Name argument must be in request.')

    # create user
    q = User.query.filter_by(name=request.args['name']).first()
    if q:
        user = q
    else:
        user = User(name=request.args['name'])
    db.session.add(user)
    db.session.flush()

    # configure roles
    roles = []
    for role in request.args.get('roles', 'default').split(','):
        q = Role.query.filter_by(name=role).first()
        if q:
            roles.append(q)
        else:
            roles.append(Role(name=role))
        db.session.add(roles[-1])
        user.roles.append(roles[-1])
        db.session.flush()

    # configure  groups
    groups = []
    for group in request.args.get('groups', 'default').split(','):
        q = Group.query.filter_by(name=group).first()
        if q:
            groups.append(q)
        else:
            groups.append(Group(name=group))
        db.session.add(groups[-1])
        user.groups.append(groups[-1])
        db.session.flush()
    db.session.commit()
    return jsonify({ 'status': 'ok' })


@app.route('/login')
def login():
    """
    Get metadata about current user.
    """
    if 'name' not in request.args:
        raise AssertionError('Name argument must be in request.')
    user = User.query.filter_by(name=request.args['name']).one()
    login_user(user)
    return jsonify({ 'status': 'ok' })


@app.route('/profile')
def profile():
    """
    Get metadata about current user.
    """
    if current_user.is_anonymous:
        return jsonify({})
    return jsonify(dict(
        name=current_user.name,
        roles=[role.name for role in current_user.roles],
        groups=[group.name for group in current_user.groups]
    ))


@app.route('/role-default')
@login_required
@authorize.has_role('default')
def authorize_role_default():
    return jsonify({ 'status': 'ok' })


@app.route('/role-admin')
@login_required
@authorize.has_role('admin')
def authorize_role_admin():
    return jsonify({ 'status': 'ok' })



# setup
# -----
@app.before_first_request
def create_tables():
    db.create_all()
    return


@login_manager.user_loader
def load_user(user_id):
    return User.query.filter_by(id=user_id).first()



# -*- coding: utf-8 -*-
#
# Testing for role- and group-based access control.
#
# ------------------------------------------------


# imports
# -------
from flask import g

from .fixtures import ArticleFactory


# jinja
# -----
class TestJinja(object):

    def test_rendering(self, client, reader, editor, restricted, admin):
        g.user = None
        ArticleFactory.create(
            name='Jinja Article',
            owner=editor,
            group=editor.groups[0]
        )

        # no content shown
        g.user = restricted
        response = client.get('/feed')
        assert response.status_code == 200
        assert 'Create Article' not in str(response.data)
        assert 'Jinja Article' not in str(response.data)

        # content shown
        g.user = reader
        response = client.get('/feed')
        assert response.status_code == 200
        assert 'Create Article' in str(response.data)
        assert 'Update Article' not in str(response.data)
        assert 'Jinja Article' in str(response.data)
        assert 'Delete Article' not in str(response.data)

        # delete button shown
        g.user = admin
        response = client.get('/feed')
        assert response.status_code == 200
        assert 'Create Article' in str(response.data)
        assert 'Update Article' not in str(response.data)
        assert 'Jinja Article' in str(response.data)
        assert 'Delete Article' in str(response.data)
        return
